/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Swarco Traffic Systems GmbH
 * Entwicklung
 * Adolf-Dambach-Strae
 * D-76571 Gaggenau
 * Phone: +49 (0)7225-64-0
 * Fax: +49 (0)7225-64-300
 * mailto: office.sts@swarco.de
 */

package de.bsvrz.ste.stealgsba.algorithmen.lib;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Diese Klasse implementiert ein Objekt vom Typ 'typ.messQuerschnitt'.<p>
 * 
 * Die Klasse dient zur Verwaltung der Messquerschnittsdaten.
 * 
 * ber Getter und Setter kann auf die Daten zugegriffen werden.
 *
 * @version $Revision: 1.5 $ / $Date: 2015/09/02 12:49:38 $ / ($Author: Asal $)
 *
 * @author Swarco Traffic Systems GmbH
 * @author Rolf Asal
 *
 */
public class MessQuerschnitt implements ClientReceiverInterface
{
	/**
   * Stellt Methoden fr die Debugausgabe zur Vefgung.
   */
	private Debug m_debug = Debug.getLogger();
	/**
	 * Aktuelle Verbindung zum Datenverteiler.
	 */
	private ClientDavConnection m_connection;
	/**
	 * Zugriff auf das aktuelle Datenmodell.
	 */
	private DataModel m_dataModel;
	/**
	 * Systemobjekt des Messquerschnitt.
	 */
	private SystemObject m_objekt;
	
	/**
	 * Attributgruppe fr den Datenempfang Kurzzeitdaten Messquerschnitt.
	 */
	private AttributeGroup m_atgVerkehrsDatenKurzZeitMq;
	/**
	 * Attributgruppe fr den Datenempfang Prognosedaten Messquerschnitt.
	 */
	private AttributeGroup m_atgVerkehrsDatenKurzZeitTrendextraPolationMq;
	/**
	 * Attributgruppe fr den Datenempfang Kurzzeitdaten Messquerschnitt.
	 */
	private AttributeGroup m_atgStoerfallZustand;
	/**
	 * Aspekt fr den Datenempfang von Analysewerte.
	 */
	private Aspect m_aspAnalyse;
	/**
	 * Aspekt fr den Datenempfang von Prognosewerte.
	 */
	private Aspect m_aspPrognose;	
	/**
	 * Aspekt fr den Datenempfang von Strfallzustand.
	 */
	private Aspect m_aspStoerfallZustand;	
	/**
	 * Anzahl der Fahrstreifen des Messquerschnitt.
	 */
	private int m_anzahlFahrstreifen;	
	/**
	 * Liste der aktuell zum Messquerschnitt gehrenden Fahrstreifenobjekte.
	 */
	private List<SystemObject> m_fahrstreifen = new ArrayList<SystemObject>();	
	/**
	 * Speicher fr die Fahrstreifenobjekte.<br>
	 * Key ist die PID des Fahrstreifen.
	 */
	private HashMap<String, FahrStreifen> m_mapFahrstreifen = new HashMap<String, FahrStreifen>();
	
	/**
	 * Zugriff auf die empfangenen Kurzzeitdaten.
	 */
	private VerkehrsDatenKurzZeitMq m_verkehrsDatenKurzZeitMq;
	/**
	 * Zugriff auf die empfangenen Prognosedaten.
	 */
	private VerkehrsDatenKurzZeitTrendExtraPolationMq m_verkehrsDatenKurzZeitTrendExtraPolationMq;	
	/**
	 * Zugriff auf den empfangenen Strfallzustand.
	 */
	private StoerfallZustand m_stoerfallZustand;	
	/**
	 * Offset des Messquerschnitt auf dem Straensegment in Zentimeter.
	 */
	private long m_offset;	
	/**
	 * Straensegment auf dem der Messquerschnitt liegt.
	 */
	private SystemObject m_strassenSegment;

	/**
	 * Konstruktor der Klasse mit bernahme der Parameter.<p>
	 * Bestimmen der Fahrstreifen und Erzeugen der Objekte fr die Verwaltung der
	 * Fahrstreifendaten.
	 * 
	 * @param connection Aktuelle Verbindung zum Datenverteiler.
	 * @param objekt Systemobjekt Messquerschnitt.
	 */
	public MessQuerschnitt(ClientDavInterface connection, SystemObject objekt)
	{
		this(connection, objekt, true, null);
	}

	/**
	 * Konstruktor der Klasse mit bernahme der Parameter.<p>
	 * Bestimmen der Fahrstreifen und Erzeugen der Objekte fr die Verwaltung der
	 * Fahrstreifendaten.
	 * 
	 * @param connection Aktuelle Verbindung zum Datenverteiler.
	 * @param objekt Systemobjekt Messquerschnitt.
	 * @param datenAnmelden Kennung Onlinedaten anmelden.
	 */
	public MessQuerschnitt(ClientDavInterface connection, SystemObject objekt, boolean datenAnmelden)
	{
		this(connection, objekt, datenAnmelden, null);
	}

	/**
	 * Konstruktor der Klasse mit bernahme der Parameter.<p>
	 * Bestimmen der Fahrstreifen und Erzeugen der Objekte fr die Verwaltung der
	 * Fahrstreifendaten.
	 * 
	 * @param connection Aktuelle Verbindung zum Datenverteiler.
	 * @param objekt Systemobjekt Messquerschnitt.
	 * @param datenAnmelden Kennung Onlinedaten anmelden.
	 * @param aspektPrognose Aspekt fr den Datenempfang Prognosewerte.
	 */
	public MessQuerschnitt(ClientDavInterface connection, SystemObject objekt, boolean datenAnmelden, Aspect aspektPrognose)
	{
		m_connection = (ClientDavConnection)connection;
		m_dataModel = m_connection.getDataModel();
		
		m_objekt = objekt;
		
		m_atgVerkehrsDatenKurzZeitMq = m_dataModel.getAttributeGroup("atg.verkehrsDatenKurzZeitMq");
		m_atgVerkehrsDatenKurzZeitTrendextraPolationMq = m_dataModel.getAttributeGroup("atg.verkehrsDatenKurzZeitTrendExtraPolationMq");
		m_atgStoerfallZustand = m_dataModel.getAttributeGroup("atg.strfallZustand");
		
		m_aspAnalyse = m_dataModel.getAspect("asp.analyse");		
		m_aspStoerfallZustand = m_dataModel.getAspect("asp.strfallVerfahrenMARZ");
		m_aspPrognose = m_dataModel.getAspect("asp.prognoseNormal");
		if (aspektPrognose != null)
			m_aspPrognose = aspektPrognose;
		
//		System.out.println("------------------> AspektProgose=" + m_aspPrognose);
		
		m_anzahlFahrstreifen = bestimmeFahrstreifen();
		
		erzeugeObjekteFahrstreifen(datenAnmelden);
		bestimmeKonfigurationsDaten();
	}
	
	/**
	 * Bestimmt aus der Konfiguration die Fahrstreifen des Messquerschnitt und legt diese
	 * in einer Liste sortiert nach der Fahrstreifenlage von rechts nach links ab.
	 * 
	 * @return Anzahl der gefundenen Fahrstreifen.
	 */
	private int bestimmeFahrstreifen()
	{
		if (m_objekt != null)
		{
			ConfigurationObject co = (ConfigurationObject)m_objekt;
			if (co != null)
			{
				HashMap<Integer, SystemObject> fsMap = new HashMap<Integer, SystemObject>();
				List<Integer> _sortedList = new ArrayList<Integer>();

				
				List<SystemObject> listeFs = co.getNonMutableSet("FahrStreifen").getElements();
				
				if (listeFs != null)
				{
					for (int i = 0; i < listeFs.size(); i++)
					{
						ConfigurationObject cofs = (ConfigurationObject)listeFs.get(i);
						if (cofs != null)
						{
							Data data = cofs.getConfigurationData(m_dataModel.getAttributeGroup("atg.fahrStreifen"));
							if (data != null)
							{
								int lage = data.getUnscaledValue("Lage").intValue();
								fsMap.put(lage, listeFs.get(i));
							}
						}
					}
					
					// Liste aufsteigend sortieren
					_sortedList.addAll(fsMap.keySet());
					Collections.sort(_sortedList);		

//					System.out.println("-----------> " + _sortedList);

					// Sortierte Ergebnisliste EreignisTyp bilden
					for (int i = 0; i < _sortedList.size(); i++)
					{
						m_fahrstreifen.add(fsMap.get(_sortedList.get(i)));
					} 

//					System.out.println(m_objekt.getPid() + " FS sortiert " + m_fahrstreifen);
				}
			}
		}
		
		return m_fahrstreifen.size();
	}

	/**
	 * Erzeugt die Objekte fr die Verwaltung der Fahrstreifendaten.
	 */
	private void erzeugeObjekteFahrstreifen(boolean datenAnmelden)
	{
		for (int i = 0; i < m_fahrstreifen.size(); i++)
		{
			SystemObject obj = m_fahrstreifen.get(i);
			
			if (!m_mapFahrstreifen.containsKey(obj.getPid()))
			{
				FahrStreifen fs = new FahrStreifen(m_connection, obj, m_aspPrognose);
				if (fs != null)
				{
					if (datenAnmelden)
					{
						fs.datenDavAnmelden();
					}
					
					m_mapFahrstreifen.put(obj.getPid(), fs);
				}
			}
		}		
	}

	/**
	 * Bestimmt die Konfigurationsdaten der Attributgruppe "atg.punktLiegtAufLinienObjekt" und schreibt
	 * sie in lokale Variable.
	 */
	private void bestimmeKonfigurationsDaten()
	{
		if (m_objekt != null)
		{
			ConfigurationObject co = (ConfigurationObject)m_objekt;
			if (co != null)
			{
				Data data = co.getConfigurationData(m_dataModel.getAttributeGroup("atg.punktLiegtAufLinienObjekt"));
				if (data != null)
				{
					m_offset = data.getUnscaledValue("Offset").longValue();
					m_strassenSegment = data.getReferenceValue("LinienReferenz").getSystemObject();
				}				
			}
		}
		
		return;
	}
	/**
	 * Anmelden der entsprechenden Daten zum Empfang am Datenverteiler.
	 */
	public void datenDavAnmelden()
	{
		empfaengerAnmeldenObjekt(this, m_objekt, m_atgVerkehrsDatenKurzZeitMq, m_aspAnalyse);
		empfaengerAnmeldenObjekt(this, m_objekt, m_atgVerkehrsDatenKurzZeitTrendextraPolationMq, m_aspPrognose);
		empfaengerAnmeldenObjekt(this, m_objekt, m_atgStoerfallZustand, m_aspStoerfallZustand);
	}
	
	/**
	 * Abmelden der angemeldeten Daten am Datenverteiler.
	 */
	public void datenDavAbmelden()
	{
		empfaengerAbmeldenObjekt(this, m_objekt, m_atgVerkehrsDatenKurzZeitMq, m_aspAnalyse);
		empfaengerAbmeldenObjekt(this, m_objekt, m_atgVerkehrsDatenKurzZeitTrendextraPolationMq, m_aspPrognose);
		empfaengerAbmeldenObjekt(this, m_objekt, m_atgStoerfallZustand, m_aspStoerfallZustand);

		for (int i = 0; i < m_fahrstreifen.size(); i++)
		{
			SystemObject obj = m_fahrstreifen.get(i);

			if (m_mapFahrstreifen.containsKey(obj.getPid()))
			{
				FahrStreifen fs = m_mapFahrstreifen.get(obj.getPid());			
				if (fs != null)
				{
					fs.datenDavAbmelden();
					
					m_mapFahrstreifen.remove(obj.getPid());
				}
			}
		}		
	}
	
	/**
   * Fr ein Objekt Attributgruppe/Aspekt Kombination am Datenverteiler anmelden.
   * 
   * @param instanz Instanz.
   * @param objekt Objekt.
   * @param atg Attributgruppe.
   * @param asp Aspekt.
   * @return true = Anmeldung erfolgreich, false = nicht erfolgreich.
   */
  private boolean empfaengerAnmeldenObjekt(ClientReceiverInterface instanz, SystemObject objekt, AttributeGroup atg, Aspect asp)
  {
    DataDescription dd;
    boolean ok = false;

    m_debug.fine(("DatenAnmeldenObjekt " + atg.getPid() + " " + asp.getPid() + " fr " + objekt));
    
    try
    {  
    	dd = new DataDescription(atg, asp);

    	m_connection.subscribeReceiver(instanz, objekt, dd, ReceiveOptions.normal(), ReceiverRole.receiver());
        
    	ok = true;
    } 
    catch (DataNotSubscribedException e)
    {
      m_debug.error(("Daten nicht zum Empfangen angemeldet " + e.getMessage()));
    } 
    
    return (ok);
  }
	/**
   * Fr ein Objekt Attributgruppe/Aspekt Kombination am Datenverteiler anmelden.
   * 
   * @param instanz Instanz.
   * @param objekt Objekt.
   * @param atg Attributgruppe.
   * @param asp Aspekt.
   * @return true = Anmeldung erfolgreich, false = nicht erfolgreich.
   */
  private boolean empfaengerAbmeldenObjekt(ClientReceiverInterface instanz, SystemObject objekt, AttributeGroup atg, Aspect asp)
  {
    DataDescription dd;
    boolean ok = false;

    m_debug.fine(("DatenAbmeldenObjekt " + atg.getPid() + " " + asp.getPid() + " fr " + objekt));
    System.out.println(("DatenAbmeldenObjekt " + atg.getPid() + " " + asp.getPid() + " fr " + objekt));
    
  	try
		{
			dd = new DataDescription(atg, asp);

			m_connection.unsubscribeReceiver(instanz, objekt, dd);
			  
			ok = true;
		}
		catch (Exception e)
		{
			m_debug.error(("Daten konnten nicht abgemeldet werden " + e.getMessage()));
		}
    
    return (ok);
  }

	/* (non-Javadoc)
	 * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
	 */
	@Override
	public void update(ResultData[] results)
	{
		for (int i = 0; i < results.length; i++)
		{
			ResultData resultData = results[i];
			
			// Kurzzeitdaten
    	if (resultData.getDataDescription().getAttributeGroup().equals(m_atgVerkehrsDatenKurzZeitMq))
    	{ 
//    		System.out.println("Kurzzeitdaten MQ -> " + resultData);

    		if (m_verkehrsDatenKurzZeitMq == null)
    			m_verkehrsDatenKurzZeitMq = new VerkehrsDatenKurzZeitMq(resultData);
    		else
    			m_verkehrsDatenKurzZeitMq.setDaten(resultData);
    		
//    		System.out.println("Kurzzeitdaten MQ -> " + resultData.getObject().getPid() + " " + m_verkehrsDatenKurzZeitMq);
    	}

    	// Prognosedaten
    	if (resultData.getDataDescription().getAttributeGroup().equals(m_atgVerkehrsDatenKurzZeitTrendextraPolationMq))
    	{ 
//    		System.out.println("Prognose MQ -> " + resultData);

    		if (m_verkehrsDatenKurzZeitTrendExtraPolationMq == null)
    			m_verkehrsDatenKurzZeitTrendExtraPolationMq = new VerkehrsDatenKurzZeitTrendExtraPolationMq(resultData);
    		else
    			m_verkehrsDatenKurzZeitTrendExtraPolationMq.setDaten(resultData);

//    		System.out.println("Prognose MQ -> " + resultData.getObject().getPid() + " " + m_verkehrsDatenKurzZeitTrendExtraPolationMq);
    	}

    	// Strfallzustand
    	if (resultData.getDataDescription().getAttributeGroup().equals(m_atgStoerfallZustand))
    	{ 
//    		System.out.println("Strfallzustand MQ -> " + resultData);

    		if (m_stoerfallZustand == null)
    			m_stoerfallZustand = new StoerfallZustand(resultData);
    		else
    			m_stoerfallZustand.setDaten(resultData);

//    		System.out.println("Strfallzustand MQ -> " + resultData.getObject().getPid() + " " + m_stoerfallZustand);
    	}
		
		}		
	}

	/**
	 * Gibt die Anzahl der Fahrstreifen des Messquerschnitts zurck.
	 * 
	 * @return Anzahl Fahrstreifen des Messquerschnitt.
	 */
	public int getAnzahlFahrstreifen()
	{
		return m_anzahlFahrstreifen;
	}

	/**
	 * Gibt die Kurzzeitdaten des Messquerschnitt zurck.
	 * 
	 * @return Kurzzeitdaten des Messquerschnitt.
	 */
	public VerkehrsDatenKurzZeitMq getVerkehrsDatenKurzZeitMq()
	{
		return m_verkehrsDatenKurzZeitMq;
	}

	/**
	 * Gibt die Prognosedaten des Messquerschnitt zurck.
	 * 
	 * @return Prognosedaten des Messquerschnitt.
	 */
	public VerkehrsDatenKurzZeitTrendExtraPolationMq getVerkehrsDatenKurzZeitTrendExtraPolationMq()
	{
		return m_verkehrsDatenKurzZeitTrendExtraPolationMq;
	}

	/**
	 * Gibt die Kurzzeitdaten des Fahrstreifen der bergebenen Pid zurck.
	 * 
	 * @param pid PID des Fahrstreifen.
	 * @return Kurzzeitdaten des Fahrstreifen oder <code>null</code>.
	 */
	public VerkehrsDatenKurzZeitFs getVerkehrsDatenKurzZeitFs(String pid)
	{
		if (pid != null)
		{
			if (m_mapFahrstreifen != null)
			{
				if (m_mapFahrstreifen.containsKey(pid))
					return m_mapFahrstreifen.get(pid).getVerkehrsDatenKurzZeitFs();
			}
		}		
		return null;
	}

	/**
	 * Gibt die Kurzzeitdaten Intervall des Fahrstreifen der bergebenen Pid zurck.
	 * 
	 * @param pid PID des Fahrstreifen.
	 * @return Kurzzeitdaten Intervall des Fahrstreifen oder <code>null</code>.
	 */
	public VerkehrsDatenKurzZeitIntervall getVerkehrsDatenKurzZeitIntervall(String pid)
	{
		if (pid != null)
		{
			if (m_mapFahrstreifen != null)
			{
				if (m_mapFahrstreifen.containsKey(pid))
					return m_mapFahrstreifen.get(pid).getVerkehrsDatenKurzZeitIntervall();
			}
		}		
		return null;
	}

	/**
	 * Gibt die Liste der Fahrstreifen zurck.
	 * 
	 * @return Liste der Fahrstreifen.
	 */
	public List<SystemObject> getFahrstreifen()
	{
		return m_fahrstreifen;
	}

	/**
	 * Diese Methode prft, ob der bergebene Fahrstreifen zu dem Messquerschnitt gehrt.
	 * 
	 * @param pidFs PID Fahrstreifen.
	 * @return true, wenn der bergebene Fahrstreifen zu dem Messquerschnitt gehrt, ansonsten false.
	 */
	public boolean gehoertFahrStreifenZuMessQuerschnitt(String pidFs)
	{
		if ((m_fahrstreifen != null) && (pidFs != null))
		{
			for (int i = 0; i < m_fahrstreifen.size(); i++)
			{
				SystemObject obj = m_fahrstreifen.get(i);
				if (obj != null)
				{
					if (obj.getPid().equals(pidFs))
						return true;
				}				
			}
		}
		
		return false;
	}
	
	/**
	 * Sucht in der Liste der Fahrstreifen den Fahrstreifen der bergebenen Lage
	 * und gibt diesen zurck.
	 * 
	 * @param lage Lage des Fahrstreifen.
	 * @return Objekt <code>FahrStreifen</code>
	 */
	public FahrStreifen getObjektFahrstreifenLage(String lage)
	{
		if (m_fahrstreifen != null)
		{
			for (int i = 0; i < m_fahrstreifen.size(); i++)
			{
				SystemObject obj = m_fahrstreifen.get(i);
				if (obj != null)
				{
					FahrStreifen fs = m_mapFahrstreifen.get(obj.getPid());
					if (fs != null)
					{
						if (fs.getLage().equals(lage))
							return fs;
					}
				}
			}
		}
		return null;
	}

	/**
	 * Sucht in der Liste der Fahrstreifen den Fahrstreifen der bergebenen PID
	 * und gibt diesen zurck.
	 * 
	 * @param pid PID des Fahrstreifen.
	 * @return Objekt <code>FahrStreifen</code>
	 */
	public FahrStreifen getObjektFahrstreifen(String pid)
	{
		if (m_fahrstreifen != null)
		{
			for (int i = 0; i < m_fahrstreifen.size(); i++)
			{
				SystemObject obj = m_fahrstreifen.get(i);
				if (obj != null)
				{
					FahrStreifen fs = m_mapFahrstreifen.get(obj.getPid());
					if (fs != null)
					{
						if (fs.getPid().equals(pid))
							return fs;
					}
				}
			}
		}
		return null;
	}

	/**
	 * Gibt die Liste mit den Instanzen 'FahrStreifen' zurck.
	 * @return Liste mit den Instanzen 'FahrStreifen'.
	 */
	public List<FahrStreifen> getInstanzFahrstreifen()
	{
		if (m_mapFahrstreifen != null)
		{
			List<FahrStreifen> lFs = new ArrayList<FahrStreifen>();
			
			Object[] oFs = m_mapFahrstreifen.keySet().toArray();
			if (oFs != null)
			{
				for (Object fs : oFs)
				{
					String _fs = (String)fs;
					lFs.add(m_mapFahrstreifen.get(_fs));
				}
			}
		
			return lFs;
		}
				
		return null;
	}
	/**
	 * Gibt das Systemobjekt des Messquerschnitts zurck.
	 * 
	 * @return Systemobjekt Messquerschnitt.
	 */
	public SystemObject getObjekt()
	{
		return m_objekt;
	}
	/**
	 * Gibt die PID des Messquerschnitts zurck.
	 * 
	 * @return PID des Messquerschnitts.
	 */
	public String getPid()
	{
		if (m_objekt != null)
		 return m_objekt.getPid();
		else
			return null;
	}

	/**
	 * Gibt den Strfallzustand zurck.
	 * @return Strfallzustand.
	 */
	public StoerfallZustand getStoerfallZustand()
	{
		return m_stoerfallZustand;
	}

	/**
	 * gibt das Straensegment auf dem der Messquerschnitt liegt zurck.
	 * @return Straensegment auf dem der Messquerschnitt liegt
	 */
	public SystemObject getStrassenSegment()
	{
		return m_strassenSegment;
	}

	/**
	 * Gibt den Offset des Messquerschnitt auf dem Straensegment in Zentimeter zurck.
	 * @return Offset des Messquerschnitt auf dem Straensegment in Zentimeter.
	 */
	public long getOffset()
	{
		return m_offset;
	}

}
