/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Swarco Traffic Systems GmbH
 * Entwicklung
 * Adolf-Dambach-Strae
 * D-76571 Gaggenau
 * Phone: +49 (0)7225-64-0
 * Fax: +49 (0)7225-64-300
 * mailto: office.sts@swarco.de
 */

package de.bsvrz.ste.stealgsba.algorithmen.lib;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse fr den Versand der Passivierungsinformation (atg.passivierungFs:asp.zustand)
 * am Objekt typ.fahrStreifen. 
 *
 * @version $Revision: 1.1 $ / $Date: 2012/07/25 12:48:38 $ / ($Author: Asal $)
 *
 * @author Swarco Traffic Systems GmbH
 * @author Rolf Asal
 *
 */
public class PassivierungFs implements ClientSenderInterface
{
	/**
   * Stellt Methoden fr die Debugausgabe zur Vefgung.
   */
	private Debug m_debug = Debug.getLogger();
	/**
	 * Aktuelle Verbindung zum Datenverteiler.
	 */
	private ClientDavConnection m_connection;
	/**
	 * Zugriff auf das aktuelle Datenmodell.
	 */
	private DataModel m_dataModel;
	/**
	 * Systemobjekt des Fahrstreifen.
	 */
	private SystemObject m_objekt;	
	/**
	 * Kennung Fahrstreifen passiviert.<br>
	 * 1 = passiviert, 0 = nicht passiviert.
	 */
	private int m_passiviert = -1;
	/**
	 * Kennung Fahrstreifen passiviert (T-1).<br>
	 * 1 = passiviert, 0 = nicht passiviert.
	 */
	private int m_passiviertAlt = -1;
	/**
	 * Attributgruppe fr den Datenversand.
	 */
	private AttributeGroup m_atgVersand;
	/**
	 * Aspekt fr den Datenversand.
	 */
	private Aspect m_aspVersand;
	/**
	 * Kennung, ob die Attributgruppenverwendung zum Versand ber den Datenverteiler angemeldet ist.
	 */
	private boolean m_datenAngemeldet = false;
	
	/**
	 * Konstruktor der Klasse.
	 * 
	 * @param connection Aktuelle Verbindung zum Datenverteiler.
	 * @param fahrstreifen Systemobjekt Fahrstreifen.
	 */
	public PassivierungFs(ClientDavInterface connection, SystemObject fahrstreifen)
	{
		m_connection = (ClientDavConnection)connection;
		m_dataModel = m_connection.getDataModel();
		
		m_objekt = fahrstreifen;
		
		m_atgVersand = m_dataModel.getAttributeGroup("atg.passivierungFs");
		m_aspVersand = m_dataModel.getAspect("asp.zustand");
	
//		quelleAnmeldenObjekt(this, m_objekt, m_atgVersand, m_aspVersand);
	}

	/**
	 * Diese Methode meldet alle erforderlichen Attributgruppenverwendungen am Datenverteier an.
	 */
	public void datenAmDavAnmelden()
	{
		if (!m_datenAngemeldet)
		{
			quelleAnmeldenObjekt(this, m_objekt, m_atgVersand, m_aspVersand);
		}
	}
	
	/**
	 * Diese Methode meldet alle erforderlichen Attributgruppenverwendungen am Datenverteier ab.
	 */
	public void datenAmDavAbmelden()
	{
		if (m_datenAngemeldet)
		{
			quelleAbmeldenObjekt(this, m_objekt, m_atgVersand, m_aspVersand);
		}
	}

	/**
	 * Setzt den Zustand der Variablen m_passiviert.
	 * 
	 * @param passiviert zu setzender Wert.
	 */
	public void setPassivierung(int passiviert)
	{
		m_passiviertAlt = m_passiviert;
		
		m_passiviert = passiviert;
	}
	
	/**
	 * Prft, ob sich der Zustand der Passivierung gendert hat.
	 * 
	 * @return true wenn nderung vorhanden, ansonsten false.
	 */
	public boolean isAenderung()
	{
		if (m_passiviert != m_passiviertAlt)
		{
			m_passiviertAlt = m_passiviert;

			return true;
		}
		
		return false;
	}
	
	/**
	 * Versendet den Zustand der Passivierung (atg.passivierungFs:asp.zustand) als Quelle
	 * ber den Datenverteiler.
	 * 
	 */
	public void sendePassivierungToDav()
	{
		if (isAenderung())
		{
			Data dataNeu = null;
			if (m_passiviert != -1)
			{
				dataNeu = m_connection.createData(m_atgVersand);
				
				dataNeu.getScaledValue("passiviert").setText((m_passiviert == 1) ? "Ja" : "Nein");
			}
			
			DataDescription dd = new DataDescription(m_atgVersand, m_aspVersand);
	    
			ResultData erg = new ResultData(m_objekt, dd, m_connection.getTime(), dataNeu);
	
			try
			{
				m_connection.sendData(erg);
			}
			catch (DataNotSubscribedException e1)
			{
				e1.printStackTrace();
			}
			catch (SendSubscriptionNotConfirmed e1)
			{
				e1.printStackTrace();
			}			
		}
	}
	
	/**
   * Fr ein Objekt Attributgruppe/Aspekt Kombination am Datenverteiler anmelden.
   * 
   * @param instanz Instanz.
   * @param objekt Objekt.
   * @param atg Attributgruppe.
   * @param asp Aspekt.
   * @return true = Anmeldung erfolgreich, false = nicht erfolgreich.
   */
  public boolean quelleAnmeldenObjekt(ClientSenderInterface instanz, SystemObject objekt, AttributeGroup atg, Aspect asp)
  {
    DataDescription dd;
    boolean ok = false;

    m_debug.fine(("QuelleAnmeldenObjekt " + atg.getPid() + " " + asp.getPid() + " fr " + objekt));
    
    try
    {  
    	dd = new DataDescription(atg, asp);

    	m_connection.subscribeSender(instanz, objekt, dd, SenderRole.source());

    	m_datenAngemeldet = true;
      
    	ok = true;
    } 
		catch (OneSubscriptionPerSendData e)
		{
			e.printStackTrace();
		} 
    
    return (ok);
  }

	/**
   * Fr ein Objekt Attributgruppe/Aspekt Kombination am Datenverteiler anmelden.
   * 
   * @param instanz Instanz.
   * @param objekt Objekt.
   * @param atg Attributgruppe.
   * @param asp Aspekt.
   * @return true = Anmeldung erfolgreich, false = nicht erfolgreich.
   */
  private void quelleAbmeldenObjekt(ClientSenderInterface instanz, SystemObject objekt, AttributeGroup atg, Aspect asp)
  {
    DataDescription dd;

    m_debug.fine(("QuelleAbmeldenObjekt " + atg + " " + asp + " fr " + objekt));
      
		dd = new DataDescription(atg, asp);

		m_connection.unsubscribeSender(instanz, objekt, dd);
    
		m_datenAngemeldet = false;
		
    return;
  }

	@Override
	public void dataRequest(SystemObject object, DataDescription dataDescription, byte state)
	{
	}

	@Override
	public boolean isRequestSupported(SystemObject object, DataDescription dataDescription)
	{
		return false;
	}

}
